# -*- coding: utf-8 -*-
#
# pygenda_version.py
# Version data in a separate file so it can be found easily by ../setup.py

# For valid/suggested versioning schemes, see:
# https://packaging.python.org/guides/distributing-packages-using-setuptools/#choosing-a-versioning-scheme

__version__ = '0.2.0'



# -*- coding: utf-8 -*-
#
# pygenda_view_year.py
# Provides the "Year View" for Pygenda.
#
# Copyright (C) 2022 Matthew Lewis
#
# This file is part of Pygenda.
#
# Pygenda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# Pygenda is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pygenda. If not, see <https://www.gnu.org/licenses/>.


from gi import require_version as gi_require_version
gi_require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GLib

import calendar
from datetime import date as dt_date, datetime as dt_datetime, timedelta
from locale import gettext as _
from icalendar import cal as iCal
from typing import Tuple

# pygenda components
from .pygenda_view import View
from .pygenda_gui import GUI, EntryDialogController
from .pygenda_config import Config
from .pygenda_calendar import Calendar
from .pygenda_util import start_end_dts_occ


# Singleton class for Year View
class View_Year(View):
	Config.set_defaults('year_view',{})
	DAY_CLASS = [ 'yearview_day_{}'.format(s) for s in ['mon','tue','wed','thu','fri','sat','sun'] ]
	GRID_COLUMNS = 37
	GRID_ROWS = 12 # One per month
	GRID_CURSOR_STYLE = 'yearview_cursor'
	ENTRY_CURSOR_STYLE = 'yearview_entry_cursor'
	_target_col = None
	_year_viewed = None
	_last_cursor = None
	_visible_occurrences = None
	_show_datecontent_pending = False
	_date_content_count = 0

	@staticmethod
	def view_name() -> str:
		# Return (localised) string to use in menu
		return _('_Year View')

	@staticmethod
	def accel_key() -> int:
		# Return (localised) keycode for menu shortcut
		k = _('year_view_accel')
		return ord(k[0]) if len(k)>0 else 0


	@classmethod
	def init(cls) -> Gtk.Widget:
		# Called on startup.
		# Gets view framework from glade file & tweaks/adds a few elements.
		# Returns widget containing view.
		cls._topbox = GUI._builder.get_object('view_year')
		cls._grid_cells = GUI._builder.get_object('year_grid_days')
		cls._date_label = GUI._builder.get_object('year_datelabel')
		cls._date_content_scroll = GUI._builder.get_object('year_datecontent_scroll')
		cls._date_content = GUI._builder.get_object('year_datecontent')
		cls._draw_day_month_labels()
		cls._year_viewed = -1 # Indicates next redraw will draw year
		cls._init_keymap()
		cls._init_grid()
		GUI._builder.get_object('year_grid_events').connect('button_press_event', cls.click_grid)
		GUI._builder.get_object('year_datecontent_events').connect('button_press_event', cls.click_events)
		return cls._topbox


	@classmethod
	def _draw_day_month_labels(cls) -> None:
		# Adds text to grid day and month labels.
		# Called only at initialisation.
		mon_names = GUI._builder.get_object('month_names')
		for i in range(cls.GRID_ROWS):
			l = Gtk.Label(calendar.month_abbr[i+1].capitalize())
			ctx = l.get_style_context()
			ctx.add_class('yearview_month_label')
			mon_names.add(l)

		st_wk = Config.get_int('global','start_week_day')
		day_names = GUI._builder.get_object('day_names')
		day_label_txt = [ d[0].capitalize() for d in calendar.day_abbr ]
		for i in range(st_wk,st_wk+cls.GRID_COLUMNS):
			l = Gtk.Label(day_label_txt[i%7])
			ctx = l.get_style_context()
			ctx.add_class('yearview_day_label')
			ctx.add_class(cls.DAY_CLASS[i%7])
			day_names.add(l)


	@classmethod
	def _init_keymap(cls) -> None:
		# Initialises KEYMAP for class. Called from init() since it needs
		# to be called after class construction, so that functions exist.
		cls._KEYMAP = {
			Gdk.KEY_Left: lambda: cls._cursor_move_lr(-1),
			Gdk.KEY_Right: lambda: cls._cursor_move_lr(1),
			Gdk.KEY_Up: lambda: cls._cursor_move_up(),
			Gdk.KEY_Down: lambda: cls._cursor_move_dn(),
			Gdk.KEY_Home: lambda: cls._cursor_move_stmon(),
			Gdk.KEY_End: lambda: cls._cursor_move_endmon(),
			Gdk.KEY_Page_Up: lambda: cls._cursor_move_pgupdn(-1),
			Gdk.KEY_Page_Down: lambda: cls._cursor_move_pgupdn(1),
			Gdk.KEY_space: lambda: cls._cursor_move_today(),
			Gdk.KEY_Return: lambda: cls.cursor_edit_entry(),
		}
		cls._KEYMAP_SHIFT = {
			Gdk.KEY_Up: lambda: cls._idxcursor_move_up(),
			Gdk.KEY_Down: lambda: cls._idxcursor_move_dn(),
		}


	@classmethod
	def _init_grid(cls) -> None:
		# Adds labels to each grid cell
		for m in range(0,cls.GRID_ROWS):
			for d in range(0,cls.GRID_COLUMNS):
				l = Gtk.Label()
				l.set_xalign(0)
				l.set_yalign(0)
				cls._grid_cells.attach(l,d,m,1,1)


	@classmethod
	def _draw_year(cls) -> None:
		# Draws year - adds formatting classes to cells, labels to some dates.
		# Called on redraw if year has changed.
		# Function a bit disorganised, tidying might be beneficial...
		c_date = GUI.cursor_date
		yr = c_date.year
		cls._year_viewed = yr
		l = GUI._builder.get_object('year_yearlabel')
		l.set_text('{:d}'.format(yr))

		st_wk = Config.get_int('global','start_week_day')
		mon_labs = GUI._builder.get_object('month_names').get_children()
		day_labs = GUI._builder.get_object('day_names').get_children()
		date = dt_date(year=yr,month=1,day=1)
		oneday = timedelta(days=1)
		today = dt_date.today()

		for m in range(1,cls.GRID_ROWS+1):
			day,daycount = calendar.monthrange(yr,m)
			col = (day-st_wk)%7 # Column for first of the month
			ctx = mon_labs[m-1].get_style_context()
			if col==0:
				ctx.add_class('yearview_leftofdaycell')
			else:
				ctx.remove_class('yearview_leftofdaycell')
			if m!=cls.GRID_ROWS:
				# This is to take account of grid lines above next month
				nday,ndaycount = calendar.monthrange(yr,m+1)
				ncol = (nday-st_wk)%7
				ndayend = ncol+ndaycount
			else:
				ncol = 7
				ndayend = 0
			for c in range(col): # Empty cells before col
				l = cls._grid_cells.get_child_at(c,m-1)
				l.set_text('')
				ctx = l.get_style_context()
				cls.remove_all_classes(ctx)
				ctx.add_class('yearview_emptycell')
				if c==col-1:
					ctx.add_class('yearview_leftofdaycell')
				if c>=ncol:
					ctx.add_class('yearview_abovedaycell')
				if m==1:
					ctx = day_labs[c].get_style_context()
					ctx.remove_class('yearview_abovedaycell')

			for d in range(daycount):
				# Potential optimisations here? Central columns are always in
				# year, so no need to remove&re-add class yearview_daycell etc.
				# Also if not start of week, will never be start of week,
				# so we know label text is already ''.
				t = ''
				if day==st_wk or d==0 or d==daycount-1:
					t = '{:d}'.format(d+1)
				l = cls._grid_cells.get_child_at(col,m-1)
				l.set_text(t)
				ctx = l.get_style_context()
				cls.remove_all_classes(ctx)
				ctx.add_class('yearview_daycell')
				ctx.add_class(cls.DAY_CLASS[day])
				if date<today:
					ctx.add_class('yearview_pastday')
				elif date==today:
					ctx.add_class('yearview_today')
				if m==1:
					ctx = day_labs[col].get_style_context()
					ctx.add_class('yearview_abovedaycell')
				day = (day+1)%7
				col += 1
				date += oneday

			for c in range(col,cls.GRID_COLUMNS): # Empty cells after
				l = cls._grid_cells.get_child_at(c,m-1)
				l.set_text('')
				ctx = l.get_style_context()
				cls.remove_all_classes(ctx)
				ctx.add_class('yearview_emptycell')
				if c<ndayend:
					ctx.add_class('yearview_abovedaycell')
				if m==1:
					ctx = day_labs[c].get_style_context()
					ctx.remove_class('yearview_abovedaycell')


	@classmethod
	def renew_display(cls) -> None:
		# Called when we switch to this view to reset state.
		cls._target_col = None


	@classmethod
	def redraw(cls, ev_changes:bool) -> None:
		# Called when redraw required.
		# ev_changes: bool indicating if events need updating too
		if cls._year_viewed != GUI.cursor_date.year:
			cls._draw_year()
			cls._last_cursor = None
			ev_changes = True
		cls._show_cursor()
		cls._show_datelabel()
		# Queue delayed redraw of day content
		if not cls._show_datecontent_pending:
			cls._date_content.foreach(Gtk.Widget.destroy)
			cls._show_datecontent_pending = True
			cls._date_content_count = 0
			# Schedule idle to add datecontent
			# Priority below draw, so datelabel will be redrawn while moving
			GLib.idle_add(cls._show_datecontent,priority=GLib.PRIORITY_HIGH_IDLE+40)
		if ev_changes:
			GLib.idle_add(cls._show_gridcontent,priority=GLib.PRIORITY_HIGH_IDLE+35)


	@classmethod
	def _show_datelabel(cls) -> None:
		# Show cursor date in label in bottom panel of view.
		# Called on redraw
		st = GUI.cursor_date.strftime(GUI.date_formatting_text_noyear)+':'
		cls._date_label.set_text(st)


	@classmethod
	def _show_datecontent(cls) -> None:
		# Show events for current cursor date in bottom panel of view.
		# Assumes that no events are currently shown.
		# Can be slow, so called in idle from redraw.
		dt = GUI.cursor_date
		cls._visible_occurrences = Calendar.occurrence_list(dt, dt+timedelta(days=1))
		r = 0
		for occ in cls._visible_occurrences:
			occ_dt_sta,occ_dt_end = start_end_dts_occ(occ)
			row = Gtk.Box()
			# Create entry mark (bullet or time) & add to row
			mark_label = cls.marker_label(occ[0], occ_dt_sta)
			ctx = mark_label.get_style_context()
			ctx.add_class('yearview_marker') # add style for CSS
			row.add(mark_label)
			# Create entry content label & add to row
			cont_label = cls.entry_text_label(occ[0],occ_dt_sta,occ_dt_end)
			cont_label.set_hexpand(True) # Also sets hexpand_set to True
			row.add(cont_label)
			cls._date_content.add(row)
			r += 1
		cls._date_content_count = r
		cls._date_content.show_all()
		cls._show_entry_cursor()
		cls._show_datecontent_pending = False


	@classmethod
	def _show_entry_cursor(cls) -> None:
		# Set style to display entry cursor (= cursor in lower section of view)
		if cls._date_content_count==0:
			# No entries for day - return
			GUI.cursor_idx_in_date = 0
			return
		i = GUI.cursor_idx_in_date
		if i<0 or i>=cls._date_content_count:
			i = cls._date_content_count-1
			GUI.cursor_idx_in_date = i
		mk = cls._date_content.get_children()[i].get_children()[0]
		ctx = mk.get_style_context()
		ctx.add_class(cls.ENTRY_CURSOR_STYLE)
		cls._last_entry_cursor = i
		cls.scroll_to_row(cls._date_content, i, cls._date_content_scroll)


	@classmethod
	def _hide_entry_cursor(cls) -> None:
		# Remove style from entry cursor (= cursor in lower section of view)
		if cls._last_entry_cursor is not None:
			mk = cls._date_content.get_children()[cls._last_entry_cursor].get_children()[0]
			ctx = mk.get_style_context()
			ctx.remove_class(cls.ENTRY_CURSOR_STYLE)
			cls._last_entry_cursor = None


	@classmethod
	def get_cursor_entry(cls) -> iCal.Event:
		# Returns entry at cursor position, or None if cursor not on entry.
		# Called from cursor_edit_entry().
		if cls._date_content_count == 0:
			return None
		return cls._visible_occurrences[GUI.cursor_idx_in_date][0]


	@classmethod
	def _show_gridcontent(cls) -> None:
		# Set styles to show events in grid.
		# Can be slow, so called in idle from redraw.
		# Potential optimisations here. Rather than getting occurences and
		# then splitting, we can add a custom calendar function to return
		# just the dates. Particularly for the repeated entries.
		yr = cls._year_viewed
		date = dt_date(year=yr,month=1,day=1)
		oneday = timedelta(days=1)
		occ_dates_single = [o[1].date() if isinstance(o[1],dt_datetime) else o[1] for o in Calendar.occurrence_list(date, dt_date(year=yr+1,month=1,day=1), include_single=True, include_repeated=False)]
		reps_list = Calendar.occurrence_list(date, dt_date(year=yr+1,month=1,day=1), include_single=False, include_repeated=True)
		occ_dates_repeated = [o[1].date() if isinstance(o[1],dt_datetime) else o[1] for o in reps_list]
		occ_dates_repeated_year = [o[1].date() if isinstance(o[1],dt_datetime) else o[1] for o in reps_list if o[0]['RRULE']['FREQ'][0]=='YEARLY']
		occ_dates_repeated_month = [o[1].date() if isinstance(o[1],dt_datetime) else o[1] for o in reps_list if o[0]['RRULE']['FREQ'][0]=='MONTHLY']
		occ_dates_repeated_week = [o[1].date() if isinstance(o[1],dt_datetime) else o[1] for o in reps_list if o[0]['RRULE']['FREQ'][0]=='WEEKLY']
		occ_dates_repeated_day = [o[1].date() if isinstance(o[1],dt_datetime) else o[1] for o in reps_list if o[0]['RRULE']['FREQ'][0]=='DAILY']
		occ_dates_repeated_hour = [o[1].date() if isinstance(o[1],dt_datetime) else o[1] for o in reps_list if o[0]['RRULE']['FREQ'][0]=='HOURLY']
		occ_dates_repeated_minute = [o[1].date() if isinstance(o[1],dt_datetime) else o[1] for o in reps_list if o[0]['RRULE']['FREQ'][0]=='MINUTELY']
		occ_dates_repeated_second = [o[1].date() if isinstance(o[1],dt_datetime) else o[1] for o in reps_list if o[0]['RRULE']['FREQ'][0]=='SECONDLY']

		for m in range(1,13):
			day,daycount = calendar.monthrange(yr,m)
			x,y = cls._date_to_cell(date)
			for d in range(daycount):
				l = cls._grid_cells.get_child_at(x,y)
				ctx = l.get_style_context()
				if date in occ_dates_single:
					ctx.add_class('yearview_entry_single')
				else:
					ctx.remove_class('yearview_entry_single')
				ctx.remove_class('yearview_entry_repeated')
				ctx.remove_class('yearview_entry_repeated_year')
				ctx.remove_class('yearview_entry_repeated_month')
				ctx.remove_class('yearview_entry_repeated_week')
				ctx.remove_class('yearview_entry_repeated_day')
				ctx.remove_class('yearview_entry_repeated_hour')
				ctx.remove_class('yearview_entry_repeated_minute')
				ctx.remove_class('yearview_entry_repeated_second')
				if date in occ_dates_repeated:
					ctx.add_class('yearview_entry_repeated')
					if date in occ_dates_repeated_year:
						ctx.add_class('yearview_entry_repeated_year')
					if date in occ_dates_repeated_month:
						ctx.add_class('yearview_entry_repeated_month')
					if date in occ_dates_repeated_week:
						ctx.add_class('yearview_entry_repeated_week')
					if date in occ_dates_repeated_day:
						ctx.add_class('yearview_entry_repeated_day')
					if date in occ_dates_repeated_hour:
						ctx.add_class('yearview_entry_repeated_hour')
					if date in occ_dates_repeated_minute:
						ctx.add_class('yearview_entry_repeated_minute')
					if date in occ_dates_repeated_second:
						ctx.add_class('yearview_entry_repeated_second')
				date += oneday
				x += 1


	@classmethod
	def _show_cursor(cls) -> None:
		# Add style class for grid cursor to the appropriate cell label
		new_coords = cls._date_to_cell(GUI.cursor_date)
		if cls._last_cursor!=new_coords:
			cls._hide_cursor()
		l = cls._grid_cells.get_child_at(*new_coords)
		ctx = l.get_style_context()
		ctx.add_class(cls.GRID_CURSOR_STYLE)
		cls._last_cursor = new_coords


	@classmethod
	def _hide_cursor(cls) -> None:
		# Remove grid cursor style class so grid cursor is not shown
		if cls._last_cursor is not None:
			l = cls._grid_cells.get_child_at(*cls._last_cursor)
			ctx = l.get_style_context()
			ctx.remove_class(cls.GRID_CURSOR_STYLE)
			cls._last_cursor = None


	@classmethod
	def _cursor_move_lr(cls, d:int) -> None:
		# Callback for user moving grid cursor left (d<0) or right (d>0).
		GUI.cursor_inc(timedelta(days=d), 0)
		cls._target_col = None
		GUI.today_toggle_date = None


	@classmethod
	def _cursor_move_up(cls) -> None:
		# Callback for user moving grid cursor up.
		cur_dt = GUI.cursor_date
		if cls._target_col is None:
			cls._target_col = cls._date_to_cell(cur_dt)[0]
		new_y = cur_dt.month-2 # months 1..12, cells 0..11
		new_yr = cur_dt.year
		if new_y<0:
			new_y = 11
			new_yr -= 1
		new_dt = cls._cell_to_date_clamped(cls._target_col, new_y, new_yr)
		GUI.cursor_set(new_dt,0)
		GUI.today_toggle_date = None


	@classmethod
	def _cursor_move_dn(cls) -> None:
		# Callback for user moving grid cursor down.
		cur_dt = GUI.cursor_date
		if cls._target_col is None:
			cls._target_col = cls._date_to_cell(cur_dt)[0]
		new_y = cur_dt.month # months 1..12, cells 0..11
		new_yr = cur_dt.year
		if new_y>11:
			new_y = 0
			new_yr += 1
		new_dt = cls._cell_to_date_clamped(cls._target_col, new_y, new_yr)
		GUI.cursor_set(new_dt,0)
		GUI.today_toggle_date = None


	@classmethod
	def _cursor_move_stmon(cls) -> None:
		# Callback for user moving grid cursor to start of month.
		new_dt = GUI.cursor_date.replace(day=1)
		GUI.cursor_set(new_dt,0)
		cls._target_col = None
		GUI.today_toggle_date = None


	@classmethod
	def _cursor_move_endmon(cls) -> None:
		# Callback for user moving grid cursor to end of month.
		dt = GUI.cursor_date
		lastday = calendar.monthrange(dt.year,dt.month)[1]
		new_dt = dt.replace(day=lastday)
		GUI.cursor_set(new_dt,0)
		cls._target_col = None
		GUI.today_toggle_date = None


	@classmethod
	def _cursor_move_pgupdn(cls, d:int) -> None:
		# Callback for user moving grid cursor pageup (d<0) or pagedown (d>0).
		cur_dt = GUI.cursor_date
		if cls._target_col is None:
			cls._target_col = cls._date_to_cell(cur_dt)[0]
		new_yr = cur_dt.year+d
		new_dt = cls._cell_to_date_clamped(cls._target_col, cur_dt.month-1, new_yr)
		GUI.cursor_set(new_dt,0)
		GUI.today_toggle_date = None


	@classmethod
	def _cursor_move_today(cls) -> None:
		# Callback for user toggling cursors between today & another date
		today = dt_date.today()
		if GUI.cursor_date != today:
			GUI.today_toggle_date = GUI.cursor_date
			GUI.today_toggle_idx = GUI.cursor_idx_in_date
			GUI.cursor_set(today,0)
			cls._target_col = None
		elif GUI.today_toggle_date is not None:
			GUI.cursor_set(GUI.today_toggle_date,GUI.today_toggle_idx)
			cls._target_col = None


	@classmethod
	def _idxcursor_move_up(cls) -> None:
		# Callback to move entry/index cursor (=cursor in bottom section) up
		if GUI.cursor_idx_in_date>0:
			GUI.cursor_idx_in_date -= 1
			cls._hide_entry_cursor()
			cls._show_entry_cursor()


	@classmethod
	def _idxcursor_move_dn(cls) -> None:
		# Callback to move entry/index cursor (=cursor in bottom section) down
		if GUI.cursor_idx_in_date<cls._date_content_count-1:
			GUI.cursor_idx_in_date += 1
			cls._hide_entry_cursor()
			cls._show_entry_cursor()


	@classmethod
	def keypress(cls, wid:Gtk.Widget, ev:Gdk.EventKey) -> None:
		# Handle key press event in Year view.
		# Called (from GUI.keypress()) on keypress (or repeat) event
		if ev.state & Gdk.ModifierType.SHIFT_MASK:
			# Shift key is being pressed
			try:
				f = cls._KEYMAP_SHIFT[ev.keyval]
				GLib.idle_add(f, priority=GLib.PRIORITY_HIGH_IDLE+30)
				return # If it gets here without errors, we're done
			except KeyError:
				pass
		try:
			f = cls._KEYMAP[ev.keyval]
			GLib.idle_add(f, priority=GLib.PRIORITY_HIGH_IDLE+30)
			return # If it gets here without errors, we're done
		except KeyError:
			pass
		if ev.state & (Gdk.ModifierType.CONTROL_MASK|Gdk.ModifierType.MOD1_MASK)==0 and Gdk.KEY_exclam <= ev.keyval <= Gdk.KEY_asciitilde:
			GLib.idle_add(EntryDialogController.newentry,chr(ev.keyval), priority=GLib.PRIORITY_HIGH_IDLE+30)


	@classmethod
	def click_grid(cls, wid:Gtk.Widget, ev:Gdk.EventButton) -> None:
		# Callback. Called whenever day grid is clicked/tapped.
		# Move grid (main) cursor to cell that was clicked.
		x = int(cls.GRID_COLUMNS*ev.x/wid.get_allocated_width())
		y = int(cls.GRID_ROWS*ev.y/wid.get_allocated_height())
		dt = cls._cell_to_date_clamped(x, y, cls._year_viewed)
		GLib.idle_add(cls._jump_to_date, dt, priority=GLib.PRIORITY_HIGH_IDLE+30)


	@classmethod
	def click_events(cls, wid:Gtk.Widget, ev:Gdk.EventButton) -> None:
		# Callback. Called whenever events area at bottom is clicked/tapped.
		# Move entry cursor to entry that was clicked
		new_idx = cls.y_to_day_row(cls._date_content, ev.y, cls._date_content_count, cls._date_content_scroll)
		if GUI.cursor_idx_in_date != new_idx:
			GUI.cursor_idx_in_date = new_idx
			cls._hide_entry_cursor()
			cls._show_entry_cursor()


	@classmethod
	def _jump_to_date(cls, dt:dt_date) -> None:
		# Idle callback to move grid cursor to given date, e.g. on click
		GUI.cursor_set(dt,0)
		cls._target_col = None


	@staticmethod
	def _date_to_cell(dt:dt_date) -> Tuple[int,int]:
		# Helper function to return cell coordinates of given date.
		stday = calendar.monthrange(dt.year,dt.month)[0] # start day of month
		x = dt.day-1+(stday-Config.get_int('global','start_week_day'))%7
		y = dt.month-1
		return x,y


	@staticmethod
	def _cell_to_date_clamped(x:int, y:int, yr:int) -> dt_date:
		# Helper function to return date given cell coordinates.
		# "Clamps" date, so if the cell is before the start or after the end
		# of the month then it returns the closest date in the month.
		m = y+1
		st_wk = Config.get_int('global','start_week_day')
		st_day,monthdays = calendar.monthrange(yr,m)
		d = x+1-(st_day-st_wk)%7
		if d<1:
			d = 1
		else:
			d = min(d,monthdays)
		return dt_date(year=yr, month=m, day=d)



# -*- coding: utf-8 -*-
#
# pygenda_widgets.py
# Date, Time and Duration entry widgets for Pygenda.
#
# Copyright (C) 2022 Matthew Lewis
#
# This file is part of Pygenda.
#
# Pygenda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# Pygenda is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pygenda. If not, see <https://www.gnu.org/licenses/>.

# To-do:
#    Add timezone field to WidgetTime (default "floating"/None).
#    Allow WidgetDuration to take values > 1 day (2d 5h 06m)
#
# !! Possible addition: ctrl+tab in WidgetDate brings up a calendar
#    (a GtkCalendar) to allow choice with touchscreen (maybe also add
#    a button to WidgetDate).
#    Maybe similar in WidgetTime if can find/make a graphical time
#    picker widget.


from gi import require_version as gi_require_version
gi_require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GLib, GObject

from datetime import date as dt_date, time as dt_time, timedelta
from calendar import monthrange
from typing import Optional

# for internationalisation/localisation
import locale
_ = locale.gettext

# pygenda components
from .pygenda_config import Config


class _WidgetDateTimeBase(Gtk.Box):
	# Base class for WidgetDate, WidgetTime, WidgetDuration.
	# Not to be used directly.
	ALLOWED_KEYS = (Gdk.KEY_BackSpace, Gdk.KEY_Delete, Gdk.KEY_Right, Gdk.KEY_Left, Gdk.KEY_Up, Gdk.KEY_Down, Gdk.KEY_Tab, Gdk.KEY_ISO_Left_Tab, Gdk.KEY_Return, Gdk.KEY_Escape)
	ALLOWED_KEYS_WITH_CTRL = (Gdk.KEY_a, Gdk.KEY_c, Gdk.KEY_x)
	SEPARATOR_KEYS = ()
	FOCUS_STYLE = 'focus'
	field_shortcuts = ()

	def __init__(self, *args, **kwds):
		super().__init__(orientation=Gtk.Orientation.HORIZONTAL, spacing=2, *args, **kwds)
		try:
			dummy = _WidgetDateTimeBase.SHORTCUT_KEYS
		except:
			# Need to set this here for it to be localisable
			_WidgetDateTimeBase.SHORTCUT_KEYS = _('ymdhm')
		self._elts = []
		self._elimpad = []
		self.set_halign(Gtk.Align.START) # default to pack from left (if LTR)


	@classmethod
	def _init_field_numeric(cls, mn:int, mx:int, zeropad:bool=True, align:float=0.5) -> Gtk.Entry:
		# Returns a new numeric field widget with the given properties.
		# Called by child classes in their initialisation.
		ln = len(str(mx))
		f = Gtk.Entry()
		f.set_max_length(ln)
		f.set_width_chars(ln)
		f.set_input_purpose(Gtk.InputPurpose.DIGITS)
		f.set_alignment(align) # default 0.5 = center digits
		f.set_activates_default(True)
		pad = ln if zeropad else 0 # Length with zero padding
		f.connect('focus-out-event', cls.validate_entry, mn, mx, pad)
		return f


	def _init_navigation(self) -> None:
		# Use self._elts array to setup navigation between subfields.
		if not Config.get_bool('global', 'tab_elts_datetime'):
			# Make Tab key only go to first elt
			self.set_focus_chain((self._elts[0],))
		self._focus_count = 0 # init counter for focus in/out
		for i in range(len(self._elts)):
			self._elts[i].connect('key-press-event', self._key_event, i)
			self._elts[i].connect('changed', self._changed)
			self._elts[i].connect('focus-out-event', self._focus_out)
			self._elts[i].connect('focus-in-event', self._focus_in)


	def _focus_in(self, entry:Gtk.Widget, ev:Gdk.EventFocus) -> None:
		# Event callback for when any sub-field gets the focus
		# This and the next function allow us to style whole widget with .focus
		# We update counter to avoid unnecessary remove/add styles
		if self._focus_count==0:
			self.get_style_context().add_class(self.FOCUS_STYLE)
		self._focus_count += 1

	def _focus_out(self, entry:Gtk.Widget, ev:Gdk.EventFocus) -> None:
		# Event callback for when any sub-field looses the focus
		# Do the work in idle so in leave/enter pairs, leave happens last
		GLib.idle_add(_WidgetDateTimeBase._do_focus_out, self)

	def _do_focus_out(self) -> None:
		# Idle event callback for when any sub-field loses focus
		self._focus_count -= 1
		if self._focus_count==0:
			self.get_style_context().remove_class(self.FOCUS_STYLE)


	def _key_event(self, entry:Gtk.Entry, ev:Gdk.EventKey, idx:int) -> bool:
		# Event callback for keypress events.
		# Return True if this handles keypress, and False to propagate event.
		try:
			txt_len = entry.get_text_length()
			c_pos = entry.get_property('cursor-position')
			sel_bnds = entry.get_selection_bounds()
			at_st = c_pos==0 or (sel_bnds and sel_bnds[0]==0)
			at_end = c_pos==txt_len or (sel_bnds and sel_bnds[1]==txt_len)
			one_from_max = not sel_bnds and c_pos==entry.get_max_length()-1
		except AttributeError:
			# Widget is not a GTkEntry (probably an am/pm combo-box)
			at_st = True
			at_end = True
			sel_bnds = False

		# Move left/right
		if at_st and ev.keyval==Gdk.KEY_Left and idx>0:
			self._elts[idx-1].grab_focus()
			return True
		if at_end and ev.keyval==Gdk.KEY_Right and idx+1<len(self._elts):
			self._elts[idx+1].grab_focus()
			return True

		# Increase/decrease keypresses
		shiftdown = ev.state&Gdk.ModifierType.SHIFT_MASK
		if ev.keyval in (Gdk.KEY_plus,Gdk.KEY_greater) or (shiftdown and ev.keyval==Gdk.KEY_Up):
			try:
				v = int(entry.get_text())
				v += 1
			except ValueError:
				# field did not contain a number (e.g. empty)
				v = self._elimpad[idx][0]
			lp = self._elimpad[idx][3]
			if v > self._elimpad[idx][1]:
				v = self._elimpad[idx][0 if lp else 1]
			pad = self._elimpad[idx][2]
			entry.set_text('{{:0{:d}d}}'.format(pad).format(v))
			entry.select_region(0,-1)
			return True
		if ev.keyval in (Gdk.KEY_minus,Gdk.KEY_less) or (shiftdown and ev.keyval==Gdk.KEY_Down):
			try:
				v = int(entry.get_text())
				v -= 1
			except ValueError:
				# field did not contain a number (e.g. empty)
				v = self._elimpad[idx][1]
			lp = self._elimpad[idx][3]
			if v < self._elimpad[idx][0]:
				v = self._elimpad[idx][1 if lp else 0]
			pad = self._elimpad[idx][2]
			entry.set_text('{{:0{:d}d}}'.format(pad).format(v))
			entry.select_region(0,-1)
			return True

		# Move to next field from typing separator
		# Small problem here. The '-' key can be a separator and can also mean
		# "decrease value". Since the latter is checked above it has priority.
		if at_end and not sel_bnds and ev.keyval in self.SEPARATOR_KEYS:
			self._elts[idx+1].grab_focus()
			return True

		# Shortcut keys to jump to fields, e.g. 'm' to jump to month field
		try:
			new_id = self.field_shortcuts.index(ev.keyval)# not found -> exceptn
			self._elts[new_id].grab_focus()
			return True
		except ValueError:
			pass

		# Type a digit
		if ev.keyval>=Gdk.KEY_0 and ev.keyval<=Gdk.KEY_9:
			if at_end and one_from_max and idx+1<len(self._elts):
				# In this case, handle insert and move to next field here
				entry.insert_text('{:d}'.format(ev.keyval-Gdk.KEY_0),-1)
				self._elts[idx+1].grab_focus()
				return True
			# Otherwise number keys can safely be handled by standard code
			return False
		# Return False to allow key to be handled if it's in "safe" list
		return ev.keyval not in (self.ALLOWED_KEYS_WITH_CTRL if ev.state==Gdk.ModifierType.CONTROL_MASK else self.ALLOWED_KEYS)


	def _changed(self, entry:Gtk.Widget) -> None:
		# If one subcomponent changes, cascade 'changed' signal down
		self.emit('changed')


	@staticmethod
	def validate_entry(entry:Gtk.Entry, ev:Gdk.EventFocus, mn:int, mx:int, pad:int) -> None:
		# Helper function to validate (and correct) date/time fields.
		# Used as callback function when focus leaves the field.
		entry.select_region(0,0) # remove highlight
		try:
			v = int(entry.get_text())
			v = max(mn,v)
			v = min(mx,v)
		except ValueError:
			v = mn
		entry.set_text('{{:0{:d}d}}'.format(pad).format(v))


	def grab_focus(self) -> None:
		# Provide implementation of widget class's grab_focus()
		self._elts[0].grab_focus()


# Allow _WidgetDateTimeBase to emit its own 'changed' signals
GObject.signal_new('changed', _WidgetDateTimeBase, GObject.SIGNAL_RUN_LAST, None, ())


class WidgetDate(_WidgetDateTimeBase):
	# Widget for date entry fields, to use (eg) in "New Entry" dialog.
	__gtype_name__ = 'WidgetDate'
	SEPARATOR_KEYS = (Gdk.KEY_slash, Gdk.KEY_minus, Gdk.KEY_period)

	def __init__(self, dt, *args, **kwds):
		super().__init__(*args, **kwds)

		self.field_year = self._init_field_numeric(1,9999,zeropad=False)
		self.field_month = self._init_field_numeric(1,12)
		self.field_day = self._init_field_numeric(1,31)

		sep = Config.get('global','date_sep')
		first = True
		self.field_shortcuts = []

		from .pygenda_gui import GUI # Delayed import to avoid circular dep
		for ch in GUI.date_order:
			if not first:
				self.pack_start(Gtk.Label(sep), False, False, 0)
			if ch=='Y':
				f = self.field_year
				lp = (1,9999,0,False) # min,max,pad,loop
				self.field_shortcuts.append(ord(self.SHORTCUT_KEYS[0]))
			elif ch=='M':
				f = self.field_month
				lp = (1,12,2,True)
				self.field_shortcuts.append(ord(self.SHORTCUT_KEYS[1]))
			else: # ch==D
				f = self.field_day
				lp = (1,31,2,True)
				self.field_shortcuts.append(ord(self.SHORTCUT_KEYS[2]))
			self._elts.append(f) # For _init_navigation()
			self._elimpad.append(lp) # For +/- keys etc
			self.pack_start(f, False, False, 0)
			first = False

		self.set_date(dt)
		self._init_navigation() # So we can navigate among elements


	def set_date(self, dt:dt_date) -> None:
		# Set widget contents.
		self.field_year.set_text(str(dt.year))
		self.field_month.set_text('{:02d}'.format(dt.month))
		self.field_day.set_text('{:02d}'.format(dt.day))


	def get_date(self) -> dt_date:
		# Get widget contents. Raises ValueError if date invalid.
		y = int(self.field_year.get_text())
		m = int(self.field_month.get_text())
		d = int(self.field_day.get_text())
		dt = dt_date(year=y, month=m, day=d)
		return dt


	def get_date_or_none(self) -> Optional[dt_date]:
		# Get widget contents. Returns None if date invalid.
		try:
			return self.get_date()
		except ValueError:
			return None


	def get_approx_date_or_none(self) -> Optional[dt_date]:
		# Get "approximate" date contents. Approximate here means try to
		# guess invalid dates, e.g. 2022-02-30 -> 2022-02-28.
		# This is for the "goto" dialog, so it doesn't force the user to
		# enter a strictly valid date.
		try:
			return self.get_date()
		except ValueError:
			pass

		# If get_date() gave an error, try harder to get values
		today = dt_date.today() # defaults if fields are empty/meaningless
		try:
			y = int(float(self.field_year.get_text()))
			if y <= 0:
				y = 1
			elif y > 9999:
				y = 9999
		except ValueError:
			y = today.year

		try:
			m = int(float(self.field_month.get_text()))
			if m <= 0:
				m = 1
			elif m > 12:
				m = 12
		except ValueError:
			m = today.month

		try:
			d = int(float(self.field_day.get_text()))
			monthdays = monthrange(y,m)[1]
			if d <= 0:
				d = 1
			elif d > monthdays:
				d = monthdays
		except ValueError:
			d = 1

		try:
			dt = dt_date(year=y, month=m, day=d)
		except ValueError:
			dt = None
		return dt


	def is_valid_date(self) -> bool:
		# Returns True if widget date is valid, False otherwise.
		try:
			self.get_date()
		except ValueError:
			return False
		return True


class WidgetTime(_WidgetDateTimeBase):
	# Widget for time entry fields, to use (eg) in "New Entry" dialog.
	__gtype_name__ = 'WidgetTime'
	SEPARATOR_KEYS = (Gdk.KEY_colon, Gdk.KEY_period)
	am_str = None # Initialise these later when we have locale config
	pm_str = None
	am_keys = None
	pm_keys = None

	def __init__(self, dt, *args, **kwds):
		super().__init__(*args, **kwds)
		self._init_ampm_locale()
		if not WidgetTime.field_shortcuts:
			WidgetTime.field_shortcuts = [ord(c) for c in (_WidgetDateTimeBase.SHORTCUT_KEYS[3:5])]

		self.is24 = Config.get_bool('global','24hr')
		if self.is24:
			self.field_hour = self._init_field_numeric(0,23)
		else:
			self.field_hour = self._init_field_numeric(1,12,zeropad=False,align=1)
			self.field_ampm = self._init_field_ampm()
		self.field_min = self._init_field_numeric(0,59)

		self.pack_start(self.field_hour, False, False, 0)
		self.pack_start(Gtk.Label(Config.get('global','time_sep')), False, False, 0)
		self.pack_start(self.field_min, False, False, 0)

		self._elts.append(self.field_hour) # For _init_navigation()
		self._elts.append(self.field_min)

		if self.is24:
			self._elimpad.append((0,23,2,True)) # For +/- keys etc
		else:
			self.pack_start(self.field_ampm, False, False, 0)
			self._elts.append(self.field_ampm)
			self._elimpad.append((1,12,0,True))
			# Add callbacks to change am/pm:
			self.field_hour.connect('key-press-event', self._remote_ampm_key_event, self.field_ampm)
			self.field_min.connect('key-press-event', self._remote_ampm_key_event, self.field_ampm)
		self._elimpad.append((0,59,2,True))

		self.set_time(dt)
		self._init_navigation() # So we can navigate among elements


	@classmethod
	def _init_ampm_locale(cls) -> None:
		# Initialise localised am/pm strings if they're None
		if cls.am_str is None:
			# Use strftime to get locale-dependent am/pm strings
			cls.am_str = dt_time().strftime('%p')
			if not cls.am_str: # For locales with "" am/pm string (eg fr_FR)
				cls.am_str = 'am' # fallback value
			cls.pm_str = dt_time(12).strftime('%p') # at 12:00
			if not cls.pm_str:
				cls.pm_str = 'pm' # fallback value

			# Now make combobox shortcut keys equal 1st chars of these strings
			# (Edge case where 1st chars of am/pm are the same?)
			ch = cls.am_str[0]
			if ch.lower() == ch.upper():
				cls.am_keys = (ord(ch),)
			else:
				cls.am_keys = (ord(ch.lower()),ord(ch.upper()))

			ch = cls.pm_str[0]
			if ch.lower() == ch.upper():
				cls.pm_keys = (ord(ch),)
			else:
				cls.pm_keys = (ord(ch.lower()),ord(ch.upper()))


	@classmethod
	def _init_field_ampm(cls) -> Gtk.ComboBoxText:
		# Returns a new am/pm field - a ComboBox.
		f = Gtk.ComboBoxText()
		f.append('a', cls.am_str)
		f.append('p', cls.pm_str)
		f.connect('key-press-event', cls._ampm_key_event)
		return f


	@staticmethod
	def _remote_ampm_key_event(entry:Gtk.Widget, ev:Gdk.EventKey, ampmfield:Gtk.ComboBoxText) -> bool:
		# Callback handler so can press a/p in other fields and have
		# it affect the am/pm field.
		if ev.keyval in WidgetTime.am_keys:
			ampmfield.set_active_id('a')
			return True
		if ev.keyval in WidgetTime.pm_keys:
			ampmfield.set_active_id('p')
			return True
		return False


	@staticmethod
	def _ampm_key_event(entry:Gtk.ComboBoxText, ev:Gdk.EventKey) -> bool:
		# Callback handler for keypresses when am/pm field is focussed.
		# Return True if this handles keypress and nothing further to do.
		if ev.keyval==Gdk.KEY_space:
			entry.popup()
			return True
		if ev.keyval in WidgetTime.am_keys:
			entry.set_active_id('a')
			return True
		if ev.keyval in WidgetTime.pm_keys:
			entry.set_active_id('p')
			return True
		shiftdown = ev.state&Gdk.ModifierType.SHIFT_MASK
		if ev.keyval in (Gdk.KEY_plus,Gdk.KEY_minus,Gdk.KEY_greater,Gdk.KEY_less) or (shiftdown and ev.keyval in (Gdk.KEY_Up,Gdk.KEY_Down)):
			entry.set_active_id('p' if entry.get_active_id()=='a' else 'a')
			return True
		if ev.keyval==Gdk.KEY_Up:
			return entry.get_toplevel().child_focus(Gtk.DirectionType.UP)
		if ev.keyval==Gdk.KEY_Down:
			return entry.get_toplevel().child_focus(Gtk.DirectionType.DOWN)
		if ev.keyval==Gdk.KEY_Return:
			dlg = entry.get_toplevel()
			if dlg:
				dlg.response(Gtk.ResponseType.OK)
				return True
		# Return False to allow key to be handled if it's in "safe" list
		return ev.keyval not in _WidgetDateTimeBase.ALLOWED_KEYS


	def set_time(self, tm:dt_time) -> None:
		# Set widget contents.
		if self.is24:
			self.field_hour.set_text('{:02d}'.format(tm.hour))
		else: # 12hr
			h = tm.hour
			self.field_ampm.set_active_id('a' if h<12 else 'p')
			if h==0:
				h = 12
			elif h>12:
				h -= 12
			self.field_hour.set_text('{:d}'.format(h))
		self.field_min.set_text('{:02d}'.format(tm.minute))


	def get_time(self) -> dt_time:
		# Get widget contents. Raises ValueError if time invalid.
		h = int(self.field_hour.get_text())
		if not self.is24:
			if h==12:
				h = 0
			if self.field_ampm.get_active_id()=='p':
				h += 12
		m = int(self.field_min.get_text())
		tm = dt_time(hour=h, minute=m)
		return tm


	def get_time_or_none(self) -> Optional[dt_time]:
		# Get widget contents. Returns None if time invalid.
		try:
			return self.get_time()
		except ValueError:
			return None


	def is_valid_time(self) -> bool:
		# Returns True if widget contents give a valid time; False otherwise.
		try:
			self.get_time()
		except ValueError:
			return False
		return True


class WidgetDuration(_WidgetDateTimeBase):
	# Widget for duration entry fields, to use (eg) in "New Entry" dialog.
	__gtype_name__ = 'WidgetDuration'
	SEPARATOR_KEYS = (Gdk.KEY_colon, Gdk.KEY_period)

	def __init__(self, timed, *args, **kwds):
		super().__init__(*args, **kwds)
		if not WidgetDuration.field_shortcuts:
			WidgetDuration.field_shortcuts = [ord(c) for c in (_WidgetDateTimeBase.SHORTCUT_KEYS[3:5])]

		# !! For now we limit to <24hrs
		self.field_hour = self._init_field_numeric(0,23,zeropad=False,align=1)
		self.field_min = self._init_field_numeric(0,59)

		self.pack_start(self.field_hour, False, False, 0)
		self.pack_start(Gtk.Label(Config.get('global','time_sep')), False, False, 0)
		self.pack_start(self.field_min, False, False, 0)

		self._elts.append(self.field_hour) # For _init_navigation()
		self._elts.append(self.field_min)
		self._elimpad.append((0,23,0,False)) # For +/- keys etc
		self._elimpad.append((0,59,2,True))

		self.set_duration(timed)
		self._init_navigation() # So we can navigate among elements


	def set_duration(self, timed:timedelta) -> None:
		# Set widget contents.
		tot_min = int(timed.total_seconds()//60)
		hr,mn = divmod(tot_min,60)
		self.field_hour.set_text('{:d}'.format(hr))
		self.field_min.set_text('{:02d}'.format(mn))


	def get_duration(self) -> timedelta:
		# Get widget contents. Raises ValueError if duration invalid.
		h = int(self.field_hour.get_text())
		m = int(self.field_min.get_text())
		if not (0<=m<60) or h<0:
			raise(ValueError)
		td = timedelta(hours=h, minutes=m)
		return td


	def get_duration_or_none(self) -> Optional[timedelta]:
		# Get widget contents. Returns None if duration invalid.
		try:
			return self.get_duration()
		except ValueError:
			return None


# CSS names
try:
	# Requires GTK 3.20+
	WidgetDate.set_css_name('date_entry')
	WidgetTime.set_css_name('time_entry')
	WidgetDuration.set_css_name('duration_entry')
except AttributeError:
	pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# pygenda/__main__.py
# Main entry point for Pygenda.
#
# Copyright (C) 2022 Matthew Lewis
#
# This file is part of Pygenda.
#
# Pygenda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# Pygenda is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pygenda. If not, see <https://www.gnu.org/licenses/>.


# Import pygenda components
from .pygenda_gui import GUI
from .pygenda_calendar import Calendar


if __name__=="__main__":
	GUI.init()
	GUI.main()



# -*- coding: utf-8 -*-
#
# pygenda_util.py
# Miscellaneous utility functions for Pygenda.
#
# Copyright (C) 2022 Matthew Lewis
#
# This file is part of Pygenda.
#
# Pygenda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# Pygenda is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pygenda. If not, see <https://www.gnu.org/licenses/>.


from calendar import day_abbr,month_abbr
from icalendar import cal as iCal
from datetime import date, time, datetime, timedelta
import locale
from typing import Tuple

from .pygenda_config import Config


def datetime_to_date(dt:date) -> date:
	# Extract date from datetime object (which might be a date)
	try:
		return dt.date()
	except AttributeError:
		return dt


def datetime_to_time(dt:date):
	# Extract time from datetime object.
	# Return False if no time component.
	try:
		return dt.time()
	except AttributeError:
		return False


def start_end_dts_event(event:iCal.Event) -> Tuple[date,date]:
	# Return start & end time of an event.
	# End time calculated from duration if needed; is None if no end/duration.
	start = event['DTSTART'].dt
	if 'DTEND' in event:
		end = event['DTEND'].dt
	elif 'DURATION' in event:
		end = start + event['DURATION'].dt
	else:
		end = None
	return start,end


def start_end_dts_occ(occ:Tuple[iCal.Event,date]) -> Tuple[date,date]:
	# Return start & end time of an occurrence (an (event,date[time]) pair)
	start = occ[1]
	if 'DTEND' in occ[0]:
		d = start - occ[0]['DTSTART'].dt
		end = occ[0]['DTEND'].dt + d
	elif 'DURATION' in occ[0]:
		end = start + occ[0]['DURATION'].dt
	else:
		end = None
	return start,end


def format_time(dt) -> str:
	# Return time as string, formatted according to app config settings.
	# Argument 'dt' can be a time or a datetime.
	if Config.get_bool('global','24hr'):
		fmt = '{hr:02d}{hs:s}{min:02d}'
		return fmt.format(hr=dt.hour, min=dt.minute, hs=Config.get('global','time_sep'))
	else:
		fmt = '{hr:d}{hs:s}{min:02d}{ampm:s}'
		return fmt.format(hr=(dt.hour-1)%12+1, min=dt.minute, ampm = 'pm' if dt.hour>11 else 'am', hs=Config.get('global','time_sep'))


def format_compact_time(dt) -> str:
	# Return time as string, according to app config settings.
	# Compact if 12hr set & minutes=0: e.g. '1pm' rather than '1:00pm'
	# Argument 'dt' can be a time or a datetime.
	if Config.get_bool('global','24hr'):
		fmt = '{hr:02d}{hs:s}{min:02d}'
		return fmt.format(hr=dt.hour, min=dt.minute, hs=Config.get('global','time_sep'))
	else:
		if dt.minute:
			fmt = '{hr:d}{hs:s}{min:02d}{ampm:s}'
		else:
			fmt = '{hr:d}{ampm:s}'
		return fmt.format(hr=(dt.hour-1)%12+1, min=dt.minute, ampm = 'pm' if dt.hour>11 else 'am', hs=Config.get('global','time_sep'))


def format_compact_date(dt:date, show_year:bool) -> str:
	# Return date as string, with abbreviated day/month names.
	if show_year:
		fmt = '{day:s} {date:d} {mon:s}, {yr:d}'
	else:
		fmt = '{day:s} {date:d} {mon:s}'
	return fmt.format(day=day_abbr[dt.weekday()], date=dt.day, mon=month_abbr[dt.month], yr=dt.year)


def format_compact_datetime(dt:datetime, show_year:bool) -> str:
	# Return datetime as string, with abbreviated day/month names, compact time.
	if show_year:
		fmt = '{day:s} {date:d} {mon:s}, {yr:d}, {tm:s}'
	else:
		fmt = '{day:s} {date:d} {mon:s}, {tm:s}'
	return fmt.format(day=day_abbr[dt.weekday()], date=dt.day, mon=month_abbr[dt.month], yr=dt.year, tm=format_compact_time(dt))


def day_in_week(dt:date) -> int:
	# Return the day number of dt in the week
	# Depends on config setting global/start_week_day
	day = dt.weekday()
	start_day = Config.get_int('global', 'start_week_day')
	return (day-start_day)%7


def start_of_week(dt:date) -> int:
	# Return first day of week containing dt
	# Depends on config setting global/start_week_day
	return dt - timedelta(days=day_in_week(dt))


def dt_lte(dt_a:date, dt_b:date) -> bool:
	# Return True if dt_a <= dt_b
	# dt_a and dt_b can be dates/datetimes independently
	return _dt_lte_common(dt_a, dt_b, True)


def dt_lt(dt_a:date, dt_b:date) -> bool:
	# Return True if dt_a < dt_b
	# dt_a and dt_b can be dates/datetimes independently
	return _dt_lte_common(dt_a, dt_b, False)


def _dt_lte_common(dt_a:date, dt_b:date, equality:bool) -> bool:
	# If one has timezone info, both need it for comparison
	try:
		if dt_a.tzinfo is None:
			if dt_b.tzinfo is not None:
				dt_a = dt_a.replace(tzinfo=dt_b.tzinfo)
		else:
			if dt_b.tzinfo is None:
				dt_b = dt_b.replace(tzinfo=dt_a.tzinfo)
	except AttributeError:
		pass
 
	# Now compare
	try:
		# These will succeed if both are dates or both are datetime
		if equality:
			return dt_a <= dt_b 
		else:
			return dt_a < dt_b
	except TypeError:
		try:
			eq = (dt_a == dt_b.date()) # try a is date
			# If eq then a is *start* of day when b occurs
			return eq or (dt_a < dt_b.date())
		except AttributeError:
			eq = (dt_a.date() == dt_b) # so b is date
			# If eq then b is *start* of day when a occurs
			return False if eq else (dt_a.date() < dt_b)
	return False # shouldn't reach here


# We want to be able to sort events/todos by datetime
def _entry_lt(self, other) -> bool:
	# Return True if start date/time of self < start date/time of other
	return dt_lt(self['DTSTART'].dt,other['DTSTART'].dt)

# Attach method to classes so it can be used to sort
iCal.Event.__lt__ = _entry_lt
iCal.Todo.__lt__ = _entry_lt


def guess_date_ord_from_locale() -> str:
	# Try to divine order of date (day/mon/yr mon/day/yr etc.) from locale.
	# Return a string like 'DMY' 'MDY' etc.
	dfmt = locale.nl_langinfo(locale.D_FMT) # E.g. '%x', '%m/%d/%Y'
	# To cope with dfmt=='%x': format a time & look at result
	st = date(year=3333,month=11,day=22).strftime(dfmt)
	ret = ''
	for ch in st:
		if ch=='3':
			if not ret or ret[-1]!='Y':
				ret += 'Y'
		elif ch=='1':
			if not ret or ret[-1]!='M':
				ret += 'M'
		elif ch=='2':
			if not ret or ret[-1]!='D':
				ret += 'D'
	return ret if len(ret)==3 else 'YMD'


def guess_date_sep_from_locale() -> str:
	# Try to divine date separator (/, -, etc.) from locale.
	dfmt = locale.nl_langinfo(locale.D_FMT)
	# To cope with dfmt=='%x': format a time & look at result
	st = date(year=1111,month=11,day=11).strftime(dfmt)
	for ch in st:
		if ch!='1':
			return ch
	return '-' # fallback


def guess_time_sep_from_locale() -> str:
	# Try to divine time separator (:, etc.) from locale.
	tfmt = locale.nl_langinfo(locale.T_FMT) # E.g. '%r', '%H:%M:%S'
	# To cope with tfmt=='%x': format a time & look at result
	st = time(hour=11,minute=11,second=11).strftime(tfmt)
	for ch in st:
		if ch!='1':
			return ch
	return ':' # fallback


def guess_date_fmt_text_from_locale() -> str:
	# Try to divine date formatting string from locale
	dtfmt = locale.nl_langinfo(locale.D_T_FMT) # includes time element
	ret = ''
	i = 0
	incl = True # including chars
	fmtch = False # formatting character
	while i < len(dtfmt):
		ch = dtfmt[i]
		if ch=='%' and not fmtch and dtfmt[i+1] in 'HIMRSTXZfklnprstz':
			incl = False # skip these until space char
		if incl:
			if fmtch and ch in 'aby':
				ch = ch.upper() # uppercase version for full day/month/year
			ret += ch
			fmtch = (ch=='%') # Boolean - next char is formatting char
		if ch==' ':
			incl = True
		i += 1
	return ret if ret else '%A, %B %-d, %Y' # fallback



# -*- coding: utf-8 -*-
#
# pygenda_view.py
# "View" class definition - base class for Week View, Year View.
# Provides default implementations of functions.
#
# Copyright (C) 2022 Matthew Lewis
#
# This file is part of Pygenda.
#
# Pygenda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# Pygenda is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pygenda. If not, see <https://www.gnu.org/licenses/>.


from gi import require_version as gi_require_version
gi_require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GLib
from gi.repository.Pango import WrapMode as PWrapMode

from icalendar import cal as iCal
from datetime import date as dt_date, timedelta

from .pygenda_gui import GUI, EntryDialogController
from .pygenda_util import datetime_to_time, datetime_to_date, format_time, format_compact_date, format_compact_time, format_compact_datetime


# Base class for pygenda Views (Week View, Year View, etc.)
class View:
	# Provide default implementations of required methods.

	@staticmethod
	def view_name() -> str:
		# Return name string to use in menu, should be localised.
		# This default should never be called, so provide dummy name.
		return 'Null View'


	@staticmethod
	def accel_key() -> int:
		# Return keycode for menu shortcut, should be localised.
		return 0


	@classmethod
	def init(cls) -> Gtk.Widget:
		# Initialise view and return Gtk widget for that view.
		# Called from GUI._init_views() on startup.
		return None


	@classmethod
	def renew_display(cls) -> None:
		# Called when we switch to this view.
		# Used to reset local state; default null implementation may be enough.
		pass


	@classmethod
	def redraw(cls, ev_changes:bool) -> None:
		# Called when redraw required
		# ev_changes: bool indicating if event display needs updating too
		pass


	@classmethod
	def keypress(cls, wid:Gtk.Widget, ev:Gdk.Event) -> None:
		# Called (from GUI.keypress()) on keypress (or repeat) event
		# Default does nothing. Derived views will override.
		pass


	@classmethod
	def cursor_edit_entry(cls) -> None:
		# Opens an entry edit dialog for the entry at the cursor,
		# or to create a new entry if the cursor is not on entry.
		# Assigned to the 'Enter' key in Week and Year views.
		en = cls.get_cursor_entry()
		if en is None:
			EntryDialogController.newentry()
		else:
			EntryDialogController.editentry(en)


	@staticmethod
	def marker_label(ev:iCal.Event, dt_st:dt_date) -> Gtk.Label:
		# Returns bullet or entry time suitable for marking entries.
		# Used to display entries in Week and Year views.
		BULLET = u'•'
		BULLET_ALLDAY = u'‣'
		BULLET_TODO = u'Ⓣ'

		lab = Gtk.Label()
		lab.set_halign(Gtk.Align.END)
		lab.set_valign(Gtk.Align.START)
		if datetime_to_time(dt_st)!=False:
			mark = format_time(dt_st)
		elif type(ev) is iCal.Todo:
			mark = BULLET_TODO
		elif 'DTEND' in ev:
			mark = BULLET_ALLDAY
		else:
			mark = BULLET
		lab.set_text(mark)

		return lab


	@staticmethod
	def entry_text_label(ev:iCal.Event, dt_st:dt_date, dt_end:dt_date) -> Gtk.Label:
		# Returns a GtkLabel with entry summary + icons as content.
		# Used by Week & Year views to display entries.
		lab = Gtk.Label()
		lab.set_line_wrap(True)
		lab.set_line_wrap_mode(PWrapMode.WORD_CHAR)
		lab.set_xalign(0)
		lab.set_yalign(0)
		endtm = View.entry_endtime(dt_st,dt_end,True)
		icons = View.entry_icons(ev,True)
		d_txt = ev['SUMMARY'] if 'SUMMARY' in ev else ''
		lab.set_text(u'{:s}{:s}{:s}'.format(d_txt,endtm,icons))
		View.add_event_styles(lab, ev)
		return lab


	@staticmethod
	def entry_icons(ev:iCal.Event, prefix_space:bool) -> str:
		# Returns string of icons for entry (repeat, alarm...)
		alarm = u'♫' # alternatives: alarm clock ⏰ (U+23F0), bell 🔔 (U+1F514)
		repeat = u'⟳'
		icons = ''
		if ev.walk('VALARM'):
			icons += alarm
		if 'RRULE' in ev:
			icons += repeat
		if prefix_space and icons:
			icons = ' {:s}'.format(icons)
		return icons


	@staticmethod
	def entry_endtime(dt_st:dt_date, dt_end:dt_date, frame_text:bool) -> str:
		# Returns endttime in format suitable for displaying in entry.
		if not dt_end or dt_st==dt_end:
			return ''
		if type(dt_st) is dt_date and type(dt_end) is dt_date:
			end_date = dt_end-timedelta(1)
			if dt_st>=end_date:
				return ''
			t_str = format_compact_date(end_date, dt_st.year!=end_date.year)
		else:
			st_date = datetime_to_date(dt_st)
			end_date = datetime_to_date(dt_end)
			if st_date == end_date:
				t_str = format_compact_time(dt_end)
			elif datetime_to_time(end_date)!=False:
				t_str = format_compact_datetime(dt_end, st_date.year!=end_date.year)
			else:
				t_str = format_compact_date(dt_end, st_date.year!=end_date.year)
		if frame_text and t_str:
			t_str = u' (→{:s})'.format(t_str)
		return t_str


	@staticmethod
	def add_event_styles(wid:Gtk.Label, ev:iCal.Event) -> None:
		# Adds formatting class to label corresponding to event status.
		# Allows entry to be formatted appropriately by CSS.
		if 'STATUS' in ev and ev['STATUS'] in ('TENTATIVE','CONFIRMED','CANCELLED'):
			ctx = wid.get_style_context()
			ctx.add_class(ev['STATUS'].lower())


	@staticmethod
	def scroll_to_row(rowbox:Gtk.Box, row:int, scroller:Gtk.ScrolledWindow) -> bool:
		# Given a box displaying entries in a scroller (such as those
		# in Week & Year views), scroll to reveal row number 'row'.
		# Note: this may need to be called after rows are rendered,
		#   so that the allocated_heights are correct
		top = rowbox.get_spacing()*row # to hold max value to show top of cell
		rows = rowbox.get_children()
		for i in range(row):
			top += rows[i].get_allocated_height()
		bot = top + rows[row].get_allocated_height()
		bot -= scroller.get_allocated_height()
		# Account for padding, margin, border
		# First get the widths from the style context
		ctx = rowbox.get_style_context()
		pad = ctx.get_padding(Gtk.StateFlags.NORMAL)
		bord = ctx.get_border(Gtk.StateFlags.NORMAL)
		marg = ctx.get_margin(Gtk.StateFlags.NORMAL)
		# Looks a bit odd to be exactly on edge, so go halfway into padding
		top += bord.top + marg.top + (pad.top+1)//2
		bot += pad.top + bord.top + marg.top + (pad.bottom+1)//2
		if bot>top: # need this if row height bigger than scroll area
			bot = top # to favour showing top, rather than bottom, of cell
		adj = scroller.get_vadjustment()
		v = adj.get_value()
		if top < v:
			adj.set_value(top)
		elif v < bot:
			adj.set_value(bot)
		# According to:
		# https://developer.gnome.org/glib/stable/glib-The-Main-Event-Loop.html#g-idle-add
		# returning False is the right thing to do for one-shot calls.
		# Not sure if this is required, but let's do it anyway.
		return False


	@staticmethod
	def y_to_day_row(rowbox:Gtk.Box, y:float, maxrow:int, scroller:Gtk.ScrolledWindow=None) -> int:
		# Helper function to convert y-coord to row index in day.
		# Used to calculate entry to jump to when entry list is clicked,
		# for example in Week or Year View.
		if maxrow <= 1:
			return 0
		rs = rowbox.get_spacing()
		# Take account of padding/border/margins/row-spacing/scroller...
		ctx = rowbox.get_style_context()
		yc = ctx.get_padding(Gtk.StateFlags.NORMAL).top
		yc += ctx.get_border(Gtk.StateFlags.NORMAL).top
		yc += ctx.get_margin(Gtk.StateFlags.NORMAL).top
		yc -= rs/2
		if scroller is not None:
			yc -= scroller.get_vadjustment().get_value()

		row = -1 # return value
		rows = rowbox.get_children()
		# accumulate height of rows until it's greater than target y
		for i in range(maxrow):
			row += 1
			yc += rows[i].get_allocated_height()
			yc += rs
			if yc > y:
				break
		return row


	@staticmethod
	def remove_all_classes(ctx:Gtk.StyleContext) -> None:
		# Helper function to remove all classes from a view context.
		# Used when redrawing Year view, might be used for other views.
		for c in ctx.list_classes():
			ctx.remove_class(c)



# -*- coding: utf-8 -*-
#
# pygenda_config.py
# Provides the Config class, to access configuration settings.
#
# Copyright (C) 2022 Matthew Lewis
#
# This file is part of Pygenda.
#
# Pygenda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# Pygenda is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pygenda. If not, see <https://www.gnu.org/licenses/>.


import argparse
from gi.repository import GLib
from pathlib import Path
import configparser
from sys import stderr
from datetime import datetime
from typing import Optional


# Singleton class to handle config from .ini file & command line
class Config:
	_cparser = configparser.RawConfigParser()
	CONFIG_DIR = GLib.get_user_config_dir() + '/pygenda'
	DEFAULT_CONFIG_FILE = CONFIG_DIR + '/pygenda.ini'
	DEFAULT_ICS_FILENAME = 'pygenda.ics' # Put this here to avoid cyclic dep.


	@classmethod
	def init(cls) -> None:
		# Initialise config - read command line, config file etc.
		# This called when file is first included (see bottom of file).
		cl_args = cls._read_arguments()
		config_file = cl_args.config
		cls.date = datetime.strptime(cl_args.date, '%Y-%m-%d').date() if cl_args.date else None
		# Now read the config file
		default_config = not config_file
		if default_config:
			config_file = cls.DEFAULT_CONFIG_FILE
			# If using default, create directory if it doesn't exist
			Path(config_file).parent.mkdir(parents=True, exist_ok=True)
		if not cls._cparser.read(config_file):
			if not default_config:
				print("Configuration file {:s} not found".format(config_file), file=stderr)
				exit(-1)
		# Store the dirname so it can be used by other components
		cls.conf_dirname = Path(config_file).parent.as_posix()

		# Read 'file' from command line, write to config store as ics_file
		if cl_args.file:
			cls.set('calendar', 'ics_file', cl_args.file)
			cls.set('calendar', 'caldav_server', None)
		if cl_args.view is not None:
			cls.set('startup','view',cl_args.view.lower())


	@classmethod
	def set_defaults(cls, sect:str, value_list:dict) -> None:
		# Called by pygena components to set defaults for config parameters.
		# This way, individual views etc control their own parameters.
		if not cls._cparser.has_section(sect):
			cls._cparser.add_section(sect)
		for key in value_list:
			if not cls._cparser.has_option(sect, key):
				cls._cparser.set(sect,key,value_list[key])


	@staticmethod
	def _read_arguments() -> argparse.Namespace:
		# Helper function to parse & return command-line arguments.
		parser = argparse.ArgumentParser(description=__doc__)
		parser.add_argument('-c', '--config', metavar='FILE', type=str, default=None, help='Config file. Default: {:s}'.format(Config.DEFAULT_CONFIG_FILE))
		parser.add_argument('-d', '--date', metavar='DATE', type=str, default=None, help='Cursor startup date (YYYY-MM-DD)')
		parser.add_argument('-f', '--file', metavar='FILE', type=str, default=None, help='Calendar file. Default: {:s} in config directory'.format(Config.DEFAULT_ICS_FILENAME))
		parser.add_argument('-v', '--view', metavar='VIEW', type=str, default=None, help='Opening view')
		return parser.parse_args()


	@classmethod
	def get(cls, section:str, option:str):
		# Generic method to get a config option.
		# Return type can be Any(?) - often str, also bool, None.
		return cls._cparser.get(section,option)


	@classmethod
	def set(cls, section:str, option:str, value) -> None:
		# Generic method to set a config option.
		# 'value' parameter can be Any type.
		if not cls._cparser.has_section(section):
			cls._cparser.add_section(section)
		cls._cparser.set(section, option, value)


	@classmethod
	def get_float(cls, section:str, option:str) -> Optional[float]:
		# Get a config option as a float.
		# Return None if the value can't be interpreted as a float.
		v = cls._cparser.get(section,option)
		if v is None or v=='':
			return None
		return cls._cparser.getfloat(section,option)


	@classmethod
	def get_int(cls, section:str, option:str) -> Optional[int]:
		# Get a config option as an int.
		# Return None if the value can't be interpreted as an int.
		v = cls._cparser.get(section,option)
		if v is None or v=='':
			return None
		return cls._cparser.getint(section,option)


	@classmethod
	def get_bool(cls, section:str, option:str) -> Optional[bool]:
		# Get a config option as a bool.
		# Return None if the value can't be interpreted as a bool.
		v = cls._cparser.get(section,option)
		if v is None or v=='':
			return None
		if isinstance(v, bool):
			return v
		return cls._cparser.getboolean(section,option)


# Setup when imported
Config.init()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Generate test file for Pygenda
#
# Copyright (C) 2022 Matthew Lewis
#
# This file is part of Pygenda.
#
# Pygenda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# Pygenda is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pygenda. If not, see <https://www.gnu.org/licenses/>.
#

print('BEGIN:VCALENDAR', end='\r\n')
print('VERSION:2.0', end='\r\n')
print('PRODID:-//Semiprime//PygendaTest//EN', end='\r\n')

for x in range(50000):
	print('BEGIN:VEVENT', end='\r\n')
	print('UID:Pygenda-{:08d}'.format(x), end='\r\n')
	print('DTSTAMP:20220101T000000Z', end='\r\n')
	print('SUMMARY:Test event {:d}'.format(x), end='\r\n')
	print('DTSTART:{:04d}{:02d}{:02d}'.format(2025-x//100, 11-(x//10)%10, 20-x%10), end='\r\n')
	#print('DURATION:PT30M', end='\r\n')
	print('END:VEVENT', end='\r\n')
print('END:VCALENDAR', end='\r\n')
